import os

import pytest

from prism._utils.version import Version
from prism._common.config import StateSettings, ROOT_EXT_WEB_URL, URL


dotenv_settings = StateSettings()
_VERSION = dotenv_settings.VERSION # Version From DotEnv
_ENV_STATE = dotenv_settings.ENV_STATE

class TestVersion:
    version_instance = Version()
    env = os.environ.get('ENV_STATE', _ENV_STATE)

    def test_instance(self):
        assert self.version_instance is not None
        assert repr(self.version_instance) == f"<Version {_VERSION}-{self.env}>"

    def test_prism__version__(self):
        import prism
        assert "\n".join(
            [
                f"Version: {_VERSION}",
                f"Environment: {self.env}",
                f"API_URL: {URL}",
                f"GUI_URL: {ROOT_EXT_WEB_URL}",
            ]) == prism.__version__

    def test_version(self):
        assert _VERSION == self.version_instance.version

    def test_environment(self):
        assert self.version_instance.environment == self.env

    def test_environment_with_env(self, monkeypatch: pytest.MonkeyPatch):
        def test_load_setup(env):
            with monkeypatch.context() as monkey:
                monkey.setenv('ENV_STATE', env)
                _, loaded_env = self.version_instance.load_setup()
                return loaded_env

        for e in ['production', 'stg', 'dev']:
            assert e == test_load_setup(e)

        # production is converted to prod
        assert 'production' == test_load_setup('production')


    def test_urls(self):
        assert 'http' in self.version_instance.api_url
        assert 'http' in self.version_instance.gui_url
        assert URL == self.version_instance.api_url
        assert ROOT_EXT_WEB_URL == self.version_instance.gui_url


import pytest

from prism._utils.validate_utils import _validate_args
from prism._utils.exceptions import PrismAuthError, PrismTypeError, PrismValueError


@_validate_args
def get_portfolio(
    universe: str,
    shownid: list
):
    return shownid


def test_shownid_param(mocker):
    mocker.patch("prism._common.const.SMValues", {
                 "apple": "prism_code1", "AAPL": "prism_code1", "prism39": "prism_code3"})

    # check type
    assert isinstance(get_portfolio('S&P 500', shownid=["prism39"]), list)

    assert get_portfolio('S&P 500', shownid=["prism39"]) == ["prism_code3"]
    assert get_portfolio('S&P 500', shownid=["PRISM39"]) == ["prism_code3"]
    assert get_portfolio('S&P 500', shownid=["AAPL", "Prism39"]) == [
        "prism_code1", "prism_code3"]
    assert get_portfolio('S&P 500', shownid=["    aapl", "prism_39"]) == [
        "prism_code1", "prism_code3"]
    assert get_portfolio('S&P 500', shownid=["a p p l e", "prism 39"]) == [
        "prism_code1", "prism_code3"]


def test_wrong_shownid_param(mocker):
    mocker.patch("prism._common.const.SMValues", {
                 "apple": "prism_code1", "AAPL": "prism_code1", "prism39": "prism_code3"})

    # wrong value
    with pytest.raises(PrismValueError):
        assert get_portfolio('S&P 500', ["aaapple"]) == ["prism_code1"]

    with pytest.raises(PrismValueError):
        assert get_portfolio('S&P 500', ["APPL"]) == ["prism_code1"]

    with pytest.raises(PrismValueError):
        assert get_portfolio('S&P 500', ["apple", "prism39", "unknown"]) == [
            "prism_code1", "prism_code1"]

    # wrong type
    with pytest.raises(PrismTypeError):
        assert get_portfolio('S&P 500', shownid=0)

    with pytest.raises(PrismTypeError):
        assert get_portfolio('S&P 500', shownid={"Apple": "AAPL"}) == [
            "prism_code1"]

    # wrong attribute
    with pytest.raises(AttributeError):
        assert get_portfolio('S&P 500', [123])

    with pytest.raises(AttributeError):
        assert get_portfolio('S&P 500', shownid=["Apple", 1234])


def test_auth_error_when_using_shownid_param(mocker):
    mocker.patch("prism._common.const.SMValues", None)

    # PrismAuthError
    with pytest.raises(PrismAuthError):
        assert get_portfolio('S&P 500', ["prism39"]) == ["prism_code3"]



from prism._utils.exceptions import PrismValueError
from ..._common.const import (
    AggregationType as _AggregationType,
    EstimatePeriodType as _PeriodType,
    CurrencyTypeWithReportTrade as _CurrencyTypeWithReportTrade
)
from .._req_builder import _list_dataitem
from ..._prismcomponent.prismcomponent import _PrismComponent, _PrismDataComponent
from ..._utils import _validate_args, _get_params


__all__ = [
    'actual',
    'consensus',
    'growth',
    'guidance',
    'recommendation',
    'revision',
    'surprise',
    'dataitems',
]


_data_category = __name__.split(".")[-1]


class _PrismEstimateComponent(_PrismDataComponent, _PrismComponent):
    _component_category_repr = _data_category

    def __init__(self, **kwargs):
        if (kwargs.get('periodtype') == 'NTM') & (kwargs.get('periodout') != 0):
            raise PrismValueError("NTM period type only takes 0 periodout.")
        super().__init__(**kwargs)

    @classmethod
    def _dataitems(cls, search: str = None, package: str = None):
        return _list_dataitem(
            datacategoryid=cls.categoryid,
            datacomponentid=cls.componentid,
            search=search,
            package=package,
        )


class actual(_PrismEstimateComponent):
    """
    | Actual financial statement result that can be compared to the consensus estimate data for a data item.
    | Default frequency is quarterly.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the value (Revenue, Expense, etc.)
        periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q',  'NTM', 'YTD', 'Non-Periodic', 'Q-SA'}
            | Estimate Period in which the financial statement results are estimated.
            | An Estimate Period can be of one of the following Period Types:

            - Annual period (A)
            - Quarterly period (Q)
            - Next twelve months (NTM)
            - Year-to-date (YTD)
            - Semi-Annual (SA)
            - Non-Periodic
            - Quarterly and Semi-Annual period (Q-SA) in quarterly standard

        currency : str, {'report', 'trade', ISO3 currency}, default 'report'
            | Desired currency for the pricing data.

            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> di = prism.estimate.actual.dataitems('revenue')
        >>> di[['dataitemid', 'dataitemname']]
           dataitemid      dataitemname
        0      200032  Revenue - Actual

        >>> rev = prism.estimate.actual(dataitemid=200032, periodtype='Q')
        >>> rev_df = rev.get_data(universe='S&P 500', startdate='2010-01-01', enddate='2015-12-31', shownid=['ticker'])
        >>> rev_df
               listingid                 date  Revenue - Actual Ticker
        0        2611056  2010-01-06 12:00:00          1822.906    FDO
        1        2631428  2010-01-06 13:01:56          1697.000    MON
        2        2626233  2010-01-07 11:00:00           913.741    LEN
        3        2602822  2010-01-07 12:30:35           987.700    STZ
        4        2590296  2010-01-07 21:01:00          1270.301   APOL
                     ...                  ...               ...    ...
        10146    2598323  2015-12-18 14:15:00          3711.000    CCL
        10147    2600608  2015-12-21 21:15:00          1219.080   CTAS
        10148    2602239  2015-12-22 12:30:00          3092.700    CAG
        10149    2638038  2015-12-22 13:30:00           722.400   PAYX
        10150    2634146  2015-12-22 21:15:00          7686.000    NKE
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        periodtype: _PeriodType,
        currency: _CurrencyTypeWithReportTrade = "report",
        package : str = None,
    ):
        if periodtype in ["NTM", None]:
            raise PrismValueError(
                f"Actual cannot take {periodtype} as periodtype.",
                valid_list=_PeriodType,
                invalids=["NTM", "None"],
            )
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable dataitems for the actual datacomponent.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns:
                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*

        Examples
        --------
            >>> di = prism.estimate.actual.dataitems('net income')
            >>> di[['dataitemid', 'dataitemname']]
            dataitemid                                       dataitemname
            0      200027  Net Income (Excl. Extraordinary Items & Good W...
            1      200028                         Net Income (GAAP) - Actual
            2      200029                     Net Income Normalized - Actual
        """
        return cls._dataitems(search=search, package=package)


class consensus(_PrismEstimateComponent):
    """
    | Consensus estimate data item.
    | Default frequency is aperiodic.


    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the value (Revenue, Expense, etc.)

        periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q',  'NTM', 'YTD', 'Non-Periodic', 'Q-SA'}
            | Estimate Period in which the financial statement results are estimated.
            | An Estimate Period can be of one of the following Period Types:

            - Annual period (A)
            - Quarterly period (Q)
            - Next twelve months (NTM)
            - Year-to-date (YTD)
            - Semi-Annual (SA)
            - Non-Periodic
            - Quarterly and Semi-Annual period (Q-SA) in quarterly standard

        periodout : int
            | Determines how far out estimate to fetch.
            | For example, inputting 0 will fetch estimate data for the current period, 1 will fetch estimate for the next period.

        currency : str, {'report', 'trade', ISO3 currency}, default 'report'
            | Desired currency for the pricing data.

            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)


    Returns
    -------
        prism._PrismComponent


    Examples
    --------
        >>> di = prism.estimate.consensus.dataitems('eps')
        >>> di[['dataitemid', 'dataitemname']]
            dataitemid                                       dataitemname
        0       200047               Cash EPS - Consensus, # of Estimates
        1       200048                         Cash EPS - Consensus, High
        2       200049                          Cash EPS - Consensus, Low
        3       200050                         Cash EPS - Consensus, Mean
        4       200051                       Cash EPS - Consensus, Median
        5       200052           Cash EPS - Consensus, Standard Deviation
        6       200119  EPS (Excl. Extraordinary Items & Good Will) - ...
        7       200120  EPS (Excl. Extraordinary Items & Good Will) - ...
        8       200121  EPS (Excl. Extraordinary Items & Good Will) - ...
        9       200122  EPS (Excl. Extraordinary Items & Good Will) - ...
        10      200123  EPS (Excl. Extraordinary Items & Good Will) - ...
        11      200124  EPS (Excl. Extraordinary Items & Good Will) - ...
        12      200125             EPS (GAAP) - Consensus, # of Estimates
        13      200126                       EPS (GAAP) - Consensus, High
        14      200127                        EPS (GAAP) - Consensus, Low
        15      200128                       EPS (GAAP) - Consensus, Mean
        16      200129                     EPS (GAAP) - Consensus, Median
        17      200130         EPS (GAAP) - Consensus, Standard Deviation
        18      200131  EPS Long-Term Growth (%) - Consensus, # of Est...
        19      200132         EPS Long-Term Growth (%) - Consensus, High
        20      200133          EPS Long-Term Growth (%) - Consensus, Low
        21      200134         EPS Long-Term Growth (%) - Consensus, Mean
        22      200135       EPS Long-Term Growth (%) - Consensus, Median
        23      200136  EPS Long-Term Growth (%) - Consensus, Standard...
        24      200137         EPS Normalized - Consensus, # of Estimates
        25      200138                   EPS Normalized - Consensus, High
        26      200139                    EPS Normalized - Consensus, Low
        27      200140                   EPS Normalized - Consensus, Mean
        28      200141                 EPS Normalized - Consensus, Median
        29      200142     EPS Normalized - Consensus, Standard Deviation

        >>> eps = prism.estimate.consensus(dataitemid=200140, periodtype='Q')
        >>> eps_df = eps.get_data(universe=1, startdate='2010-01-01', enddate='2015-12-31', shownid=['ticker'])
        >>> eps_df
              listingid                 date  EPS Normalized - Consensus, Mean   Ticker
        0      20108958  2010-01-26 01:37:00                         8588.6960  A077970
        1      20108958  2011-08-29 11:56:36                         -776.0000  A077970
        2      20108958  2012-04-02 11:30:54                         2320.0000  A077970
        3      20108958  2012-08-29 01:17:12                         1200.0000  A077970
        4      20113302  2010-01-11 23:22:00                          922.0000  A078930
        ...         ...                  ...                               ...      ...
        8125  278631846  2015-04-09 00:30:47                          423.0076  A028260
        8126  278631846  2015-04-29 10:53:38                          243.5500  A028260
        8127  278631846  2015-04-29 13:30:47                          369.1384  A028260
        8128  278631846  2015-10-21 09:36:10                        14533.4700  A028260
        8129  278631846  2015-11-04 12:51:57                          497.4083  A028260
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        periodtype: _PeriodType = None,
        periodout: int = 0,
        currency: _CurrencyTypeWithReportTrade = "report",
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the estimate consensus data component.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns:
                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*


        Examples
        --------
            >>> di = prism.estimate.consensus.dataitems('net income')
            >>> di[['dataitemid', 'dataitemname']]
                dataitemid                                       dataitemname
            0       200203  Net Income (Excl. Extraordinary Items & Good W...
            1       200204  Net Income (Excl. Extraordinary Items & Good W...
            2       200205  Net Income (Excl. Extraordinary Items & Good W...
            3       200206  Net Income (Excl. Extraordinary Items & Good W...
            4       200207  Net Income (Excl. Extraordinary Items & Good W...
            5       200208  Net Income (Excl. Extraordinary Items & Good W...
            6       200209      Net Income (GAAP) - Consensus, # of Estimates
            7       200210                Net Income (GAAP) - Consensus, High
            8       200211                 Net Income (GAAP) - Consensus, Low
            9       200212                Net Income (GAAP) - Consensus, Mean
            10      200213              Net Income (GAAP) - Consensus, Median
            11      200214  Net Income (GAAP) - Consensus, Standard Deviation
            12      200215  Net Income Normalized - Consensus, # of Estimates
            13      200216            Net Income Normalized - Consensus, High
            14      200217             Net Income Normalized - Consensus, Low
            15      200218            Net Income Normalized - Consensus, Mean
            16      200219          Net Income Normalized - Consensus, Median
            17      200220  Net Income Normalized - Consensus, Standard De...
        """
        return cls._dataitems(search=search, package=package)


class growth(_PrismEstimateComponent):
    """
    | Differences in consensus estimate and the actual for a data item.
    | Default frequency is aperiodic.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the value (Revenue, Expense, etc.)
        periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q',  'NTM', 'YTD', 'Non-Periodic', 'Q-SA'}
            | Estimate Period in which the financial statement results are estimated.
            | An Estimate Period can be of one of the following Period Types:

            - Annual period (A)
            - Quarterly period (Q)
            - Next twelve months (NTM)
            - Year-to-date (YTD)
            - Semi-Annual (SA)
            - Non-Periodic
            - Quarterly and Semi-Annual period (Q-SA) in quarterly standard

        periodout : int
            | Determines how far out estimate to fetch.
            | For example, inputting 0 will fetch estimate data for the current period, 1 will fetch estimate for the next period.

        currency : str, {'report', 'trade', ISO3 currency}, default 'report'
            | Desired currency for the pricing data.

            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> di = prism.estimate.growth.dataitems()
        >>> di[['dataitemid', 'dataitemname']]
            dataitemid                                       dataitemname
        0       200276  Capital Expenditure - Growth, Quarter on Quart...
        1       200277     Cash Flow / Share - Growth, Annual, 1 Year (%)
        2       200278     Cash Flow / Share - Growth, Annual, 2 Year (%)
        3       200279  Cash Flow / Share - Growth, Quarter on Quarter...
        4       200280     Cash Flow / Share - Growth, Year over Year (%)
        5       200281                   DPS - Growth, Annual, 1 Year (%)
        6       200282                   DPS - Growth, Annual, 2 Year (%)
        7       200283               DPS - Growth, Quarter on Quarter (%)
        8       200284                   DPS - Growth, Year over Year (%)
        9       200285                EBITDA - Growth, Annual, 1 Year (%)
        10      200286                EBITDA - Growth, Annual, 2 Year (%)
        11      200287            EBITDA - Growth, Quarter on Quarter (%)
        12      200288                EBITDA - Growth, Year over Year (%)
        13      200289               Revenue - Growth, Annual, 1 Year (%)
        14      200290               Revenue - Growth, Annual, 2 Year (%)
        15      200291           Revenue - Growth, Quarter on Quarter (%)
        16      200292               Revenue - Growth, Year over Year (%)

        >>> rev_growth = prism.estimate.growth(dataitemid=200281, periodtype='Q')
        >>> rev_growth_df = rev_growth.get_data(universe='S&P 500', startdate='2010-01-01', enddate='2015-12-31', shownid=['ticker'])
        >>> rev_growth_df
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        periodtype: _PeriodType = None,
        periodout: int = 0,
        currency: _CurrencyTypeWithReportTrade = "report",
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the growth data component.


        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.


        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns:
                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*

        Examples
        --------
            >>> di = prism.estimate.growth.dataitems()
            >>> di[['dataitemid', 'dataitemname']]
                dataitemid                                       dataitemname
            0       200276  Capital Expenditure - Growth, Quarter on Quart...
            1       200277     Cash Flow / Share - Growth, Annual, 1 Year (%)
            2       200278     Cash Flow / Share - Growth, Annual, 2 Year (%)
            3       200279  Cash Flow / Share - Growth, Quarter on Quarter...
            4       200280     Cash Flow / Share - Growth, Year over Year (%)
            5       200281                   DPS - Growth, Annual, 1 Year (%)
            6       200282                   DPS - Growth, Annual, 2 Year (%)
            7       200283               DPS - Growth, Quarter on Quarter (%)
            8       200284                   DPS - Growth, Year over Year (%)
            9       200285                EBITDA - Growth, Annual, 1 Year (%)
            10      200286                EBITDA - Growth, Annual, 2 Year (%)
            11      200287            EBITDA - Growth, Quarter on Quarter (%)
            12      200288                EBITDA - Growth, Year over Year (%)
            13      200289               Revenue - Growth, Annual, 1 Year (%)
            14      200290               Revenue - Growth, Annual, 2 Year (%)
            15      200291           Revenue - Growth, Quarter on Quarter (%)
            16      200292               Revenue - Growth, Year over Year (%)
        """
        return cls._dataitems(search=search, package=package)


class recommendation(_PrismEstimateComponent):
    """
    | Recommnedation Data for a data item.
    | Default frequency is quarterly.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the value (Score, Buy, etc.)

        periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q',  'NTM', 'YTD', 'Non-Periodic', 'Q-SA'}
            | Estimate Period in which the financial statement results are estimated.
            | An Estimate Period can be of one of the following Period Types:

            - Annual period (A)
            - Quarterly period (Q)
            - Next twelve months (NTM)
            - Year-to-date (YTD)
            - Semi-Annual (SA)
            - Non-Periodic
            - Quarterly and Semi-Annual period (Q-SA) in quarterly standard

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> df = prism.estimate.recommendation.dataitems('Score')
        >>> df[['dataitemid', 'dataitemname']]
           dataitemid                     dataitemname
        0      200631           Recommendation - Score
        1      200637  Industry Recommendation - Score
        2      601998           Recommendation - Score
        >>> rec = prism.estimate.recommendation(dataitemid=200631, periodtype='LTM')
        >>> rec.get_data("LSE", "2022-01-01", shownid=['companyname'])
               listingid                 date  Recommendation - Score                     Company Name
        0       22033716  2022-01-01 00:30:00                 2.42105  The Berkeley Group Holdings plc
        1       20131190  2022-01-01 03:29:13                 1.42105         Barratt Developments plc
        2       20175571  2022-01-01 03:29:13                 2.09091                    Elementis plc
        3       25117249  2022-01-01 03:29:13                 2.06250      Lancashire Holdings Limited
        4       29847692  2022-01-01 03:29:13                 2.50000                       Hiscox Ltd
        ...          ...                  ...                     ...                              ...
        63519   20157229  2023-07-31 13:59:44                 2.00000                     RS Group plc
        63520  224316065  2023-07-31 14:31:00                 1.77778                          WPP plc
        63521   20176224  2023-07-31 14:40:11                 2.14286                     Spectris plc
        63522   20157229  2023-07-31 15:11:00                 2.00000                     RS Group plc
        63523   20182531  2023-07-31 17:04:13                 1.76923                      Pearson plc
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        periodtype: _PeriodType = None,
        package: str = None,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the recommendation data component.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns:
                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*

        Examples
        --------
            >>> di = prism.estimate.recommendation.dataitems()
            >>> di[['dataitemid', 'dataitemname']]
            dataitemid                                         dataitemname
            0      200625    Recommendation - # of Analysts Buy Recommendation
            1      200626    Recommendation - # of Analysts Hold Recommenda...
            2      200627    Recommendation - # of Analysts No Opinion Reco...
            3      200628    Recommendation - # of Analysts Outperform Reco...
            4      200629    Recommendation - # of Analysts Sell Recommenda...
            5      200630    Recommendation - # of Analysts Underperform Re...
            6      200631    Recommendation - Score
            7      200632    Industry Recommendation - # of Analysts Buy Re...
            8      200633    Industry Recommendation - # of Analysts Hold R...
            9      200634    Industry Recommendation - # of Analysts Outper...
            10     200635    Industry Recommendation - # of Analysts Sell R...
            11     200636    Industry Recommendation - # of Analysts Underp...
            12     200637    Industry Recommendation - Score
            13     601991    Recommendation - # of Analysts Buy Recommendation
            14     601993    Recommendation - # of Analysts Hold Recommenda...
            15     601995    Recommendation - # of Analysts Sell Recommenda...
            16     601996    Recommendation - # of Analysts No Opinion Reco...
            17     601997    Recommendation - # of Analysts Total Recommend...
            18     601998    Recommendation - Score
            19     602197    Recommendation - # of Analysts Overweight Reco...
            20     602198    Recommendation - # of Analysts Underweight Rec...
        """
        return cls._dataitems(search=search, package=package)


class guidance(_PrismEstimateComponent):
    """
    | Company guidance data for a data item.
    | Default frequency is aperiodic.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the value (Revenue, Expense, etc.)

        periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q',  'NTM', 'YTD', 'Non-Periodic', 'Q-SA'}
            | Estimate Period in which the financial statement results are estimated.
            | An Estimate Period can be of one of the following Period Types:

            - Annual period (A)
            - Quarterly period (Q)
            - Next twelve months (NTM)
            - Year-to-date (YTD)
            - Semi-Annual (SA)
            - Non-Periodic
            - Quarterly and Semi-Annual period (Q-SA) in quarterly standard

        periodout : int
            | Determines how far out estimate to fetch.
            | For example, inputting 0 will fetch estimate data for the current period, 1 will fetch estimate for the next period.

        currency : str, {'report', 'trade', ISO3 currency}, default 'report'
            | Desired currency for the pricing data.

            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> di = prism.estimate.guidance.dataitems('eps')
        >>> di[['dataitemid', 'dataitemname']]
            dataitemid                                       dataitemname
        0       200299                          Cash EPS - Guidance, High
        1       200300                           Cash EPS - Guidance, Low
        2       200301                           Cash EPS - Guidance, Mid
        3       200335  EPS (Excl. Extraordinary Items & Good Will) - ...
        4       200336  EPS (Excl. Extraordinary Items & Good Will) - ...
        5       200337  EPS (Excl. Extraordinary Items & Good Will) - ...
        6       200338                        EPS (GAAP) - Guidance, High
        7       200339                         EPS (GAAP) - Guidance, Low
        8       200340                         EPS (GAAP) - Guidance, Mid
        9       200341                    EPS Normalized - Guidance, High
        10      200342                     EPS Normalized - Guidance, Low
        11      200343                     EPS Normalized - Guidance, Mid

        >>> eps = prism.estimate.guidance(dataitemid=200343, periodtype='Q')
        >>> eps_df = eps.get_data(universe='S&P 500', startdate='2010-01-01', enddate='2015-12-31', shownid=['ticker'])
        >>> eps_df
             listingid                 date  EPS Normalized - Guidance, Mid  Ticker
        0      2586546  2013-11-05 21:44:00                            0.45     ANF
        1      2587881  2015-04-08 10:44:00                            0.78      AA
        2      2588077  2010-10-15 11:48:00                            0.05     ATI
        3      2588150  2014-02-06 12:30:00                            2.70     ADS
        4      2588150  2014-04-17 11:30:00                            2.70     ADS
        ..         ...                  ...                             ...     ...
        218   46639516  2015-11-03 03:31:00                            0.82    ATVI
        219   52984459  2011-10-27 13:30:00                            0.51     MJN
        220   52984459  2015-07-14 21:36:00                            0.76     MJN
        221  141637207  2015-07-13 20:30:00                            0.43     XYL
        222  262300822  2015-07-13 20:18:00                            0.40    NAVI
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        periodtype: _PeriodType = None,
        periodout: int = 0,
        currency: _CurrencyTypeWithReportTrade = "report",
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the guidance data component.


        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.


        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns:
                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*

        Examples
        --------
            >>> di = prism.estimate.guidance.dataitems('net income')
            >>> di[['dataitemid', 'dataitemname']]
            dataitemid                                       dataitemname
            0      200368  Net Income (Excl. Extraordinary Items & Good W...
            1      200369  Net Income (Excl. Extraordinary Items & Good W...
            2      200370  Net Income (Excl. Extraordinary Items & Good W...
            3      200371                 Net Income (GAAP) - Guidance, High
            4      200372                  Net Income (GAAP) - Guidance, Low
            5      200373                  Net Income (GAAP) - Guidance, Mid
            6      200374             Net Income Normalized - Guidance, High
            7      200375              Net Income Normalized - Guidance, Low
            8      200376              Net Income Normalized - Guidance, Mid
        """
        return cls._dataitems(search=search, package=package)


class revision(_PrismEstimateComponent):
    """
    | Revision in consensus estimate a data item.
    | Default frequency is aperiodic.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the value (Revenue, Expense, etc.)

        periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q',  'NTM', 'YTD', 'Non-Periodic', 'Q-SA'}
            | Estimate Period in which the financial statement results are estimated.
            | An Estimate Period can be of one of the following Period Types:

            - Annual period (A)
            - Quarterly period (Q)
            - Next twelve months (NTM)
            - Year-to-date (YTD)
            - Semi-Annual (SA)
            - Non-Periodic
            - Quarterly and Semi-Annual period (Q-SA) in quarterly standard

        periodout : int
            | Determines how far out estimate to fetch.
            | For example, inputting 0 will fetch estimate data for the current period, 1 will fetch estimate for the next period.

        aggregation : str, {'1 day', '1 week', '1 month', '2 month', '3 month', '3 month latest'}, default '1 day'
            | Aggregation time periods covered in the revisions calculations.

            .. admonition:: Warning
                :class: warning

                | If the input is **'1 week'** the resulting revision Data Component will contain data values that are sum of revision within 1 week of the data dates.
                |
                | If the input is  **'3 month latest'**, it will only account for latest revision of the same analyst. Therefore, if a same analyst revises his/her estimate more than one time within the 3 month period, only one will be counted towards calculating the revision number

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> di = prism.estimate.revision.dataitems('eps')
        >>> di[['dataitemid', 'dataitemname']]
            dataitemid                                       dataitemname
        0       200408                 Cash EPS - Revision, # of Analysts
        1       200409                          Cash EPS - Revision, Down
        2       200410                     Cash EPS - Revision, No Change
        3       200411      Cash EPS - Revision, Unfiltered # of Analysts
        4       200412               Cash EPS - Revision, Unfiltered Down
        5       200413          Cash EPS - Revision, Unfiltered No Change
        6       200414                 Cash EPS - Revision, Unfiltered Up
        7       200415                            Cash EPS - Revision, Up
        8       200472  EPS (Excl. Extraordinary Items & Good Will) - ...
        9       200473  EPS (Excl. Extraordinary Items & Good Will) - ...
        10      200474  EPS (Excl. Extraordinary Items & Good Will) - ...
        11      200475  EPS (Excl. Extraordinary Items & Good Will) - ...
        12      200476  EPS (Excl. Extraordinary Items & Good Will) - ...
        13      200477  EPS (Excl. Extraordinary Items & Good Will) - ...
        14      200478  EPS (Excl. Extraordinary Items & Good Will) - ...
        15      200479  EPS (Excl. Extraordinary Items & Good Will) - ...
        16      200480               EPS (GAAP) - Revision, # of Analysts
        17      200481                        EPS (GAAP) - Revision, Down
        18      200482                   EPS (GAAP) - Revision, No Change
        19      200483    EPS (GAAP) - Revision, Unfiltered # of Analysts
        20      200484             EPS (GAAP) - Revision, Unfiltered Down
        21      200485        EPS (GAAP) - Revision, Unfiltered No Change
        22      200486               EPS (GAAP) - Revision, Unfiltered Up
        23      200487                          EPS (GAAP) - Revision, Up
        24      200488           EPS Normalized - Revision, # of Analysts
        25      200489                    EPS Normalized - Revision, Down
        26      200490               EPS Normalized - Revision, No Change
        27      200491  EPS Normalized - Revision, Unfiltered # of Ana...
        28      200492         EPS Normalized - Revision, Unfiltered Down
        29      200493    EPS Normalized - Revision, Unfiltered No Change
        30      200494           EPS Normalized - Revision, Unfiltered Up
        31      200495                      EPS Normalized - Revision, Up

        >>> eps = prism.estimate.revision(dataitemid=200495, periodtype='Q')
        >>> eps_df = eps.get_data(universe='S&P 500', startdate='2010-01-01', enddate='2015-12-31', shownid=['ticker'])
        >>> eps_df
                listingid        date   EPS Normalized - Revision, Up  Ticker
        0         2586086  2010-04-20                               1     AFL
        1         2586086  2010-04-21                               1     AFL
        2         2586086  2010-04-22                               1     AFL
        3         2586086  2010-04-29                               2     AFL
        4         2586086  2010-05-01                               1     AFL
        ...           ...         ...                             ...     ...
        126320  306093931  2015-12-05                               1    PYPL
        126321  313064514  2015-11-24                               0     HPE
        126322  313064514  2015-11-26                               3     HPE
        126323  313064514  2015-11-27                               0     HPE
        126324  313064514  2015-12-11                               0     HPE
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        periodtype: _PeriodType = None,
        periodout: int = 0,
        aggregation: _AggregationType = '1 day',
        package: str = None,
    ):
        if periodtype in ["NTM"]:
            raise PrismValueError(
                f"Revision cannot take {periodtype} as periodtype.",
                valid_list=_PeriodType,
                invalids=["NTM"],
            )
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the revision data component.


        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.
            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns:
                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*

        Examples
        --------
            >>> di = prism.estimate.revision.dataitems('net income')
            >>> di[['dataitemid', 'dataitemname']]
                dataitemid                                       dataitemname
            0       200528  Net Income (Excl. Extraordinary Items & Good W...
            1       200529  Net Income (Excl. Extraordinary Items & Good W...
            2       200530  Net Income (Excl. Extraordinary Items & Good W...
            3       200531  Net Income (Excl. Extraordinary Items & Good W...
            4       200532  Net Income (Excl. Extraordinary Items & Good W...
            5       200533  Net Income (Excl. Extraordinary Items & Good W...
            6       200534  Net Income (Excl. Extraordinary Items & Good W...
            7       200535  Net Income (Excl. Extraordinary Items & Good W...
            8       200536        Net Income (GAAP) - Revision, # of Analysts
            9       200537                 Net Income (GAAP) - Revision, Down
            10      200538            Net Income (GAAP) - Revision, No Change
            11      200539  Net Income (GAAP) - Revision, Unfiltered # of ...
            12      200540      Net Income (GAAP) - Revision, Unfiltered Down
            13      200541  Net Income (GAAP) - Revision, Unfiltered No Ch...
            14      200542        Net Income (GAAP) - Revision, Unfiltered Up
            15      200543                   Net Income (GAAP) - Revision, Up
            16      200544    Net Income Normalized - Revision, # of Analysts
            17      200545             Net Income Normalized - Revision, Down
            18      200546        Net Income Normalized - Revision, No Change
            19      200547  Net Income Normalized - Revision, Unfiltered #...
            20      200548  Net Income Normalized - Revision, Unfiltered Down
            21      200549  Net Income Normalized - Revision, Unfiltered N...
            22      200550    Net Income Normalized - Revision, Unfiltered Up
            23      200551               Net Income Normalized - Revision, Up
        """
        return cls._dataitems(search=search, package=package)


class surprise(_PrismEstimateComponent):
    """
    | Differences in consensus estimate and the actual for a data item.
    | Default frequency is quarterly.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the value (Revenue, Expense, etc.)

        periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q',  'NTM', 'YTD', 'Non-Periodic', 'Q-SA'}
            | Estimate Period in which the financial statement results are estimated.
            | An Estimate Period can be of one of the following Period Types:

            - Annual period (A)
            - Quarterly period (Q)
            - Next twelve months (NTM)
            - Year-to-date (YTD)
            - Semi-Annual (SA)
            - Non-Periodic
            - Quarterly and Semi-Annual period (Q-SA) in quarterly standard

        currency : str, {'report', 'trade', ISO3 currency}, default 'report'
            | Desired currency for the pricing data.

            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> di = prism.estimate.surprise.dataitems('revenue')
        >>> di[['dataitemid', 'dataitemname']]
           dataitemid                   dataitemname
        0      200618  Revenue - Surpise, Difference
        1      200619     Revenue - Surpise, Percent

        >>> rev = prism.estimate.surprise(dataitemid=200618, periodtype='Q')
        >>> rev_df = rev.get_data(universe='S&P 500', startdate='2010-01-01', enddate='2015-12-31', shownid=['ticker'])
        >>> rev_df
               listingid      period        date  EPS Normalized - Surpise, Difference  Ticker
        0        2586086  2009-12-31  2010-02-02                                  0.01     AFL
        1        2586086  2010-03-31  2010-04-27                                  0.05     AFL
        2        2586086  2010-06-30  2010-07-27                                  0.01     AFL
        3        2586086  2010-09-30  2010-10-26                                  0.04     AFL
        4        2586086  2010-12-31  2011-02-01                                 -0.01     AFL
          ...          ...         ...         ...                                   ...     ...
        11516  312940617  2014-12-31  2015-01-29                                 -0.25    GOOG
        11517  312940617  2015-03-31  2015-04-23                                 -0.04  GOOC.V
        11518  312940617  2015-03-31  2015-04-23                                 -0.04    GOOG
        11519  312940617  2015-06-30  2015-07-16                                  0.30  GOOC.V
        11520  312940617  2015-06-30  2015-07-16                                  0.30    GOOG
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        periodtype: _PeriodType,
        currency: _CurrencyTypeWithReportTrade = "report",
        package : str = None,
    ):
        if periodtype in ["NTM", None]:
            raise PrismValueError(
                f"Surprise cannot take {periodtype} as periodtype.",
                valid_list=_PeriodType,
                invalids=["NTM", "None"],
            )
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the surprise data component.


        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns:
                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*

        Examples
        --------
            >>> di = prism.estimate.surprise.dataitems('net income')
            >>> di[['dataitemid', 'dataitemname']]
            dataitemid                                       dataitemname
            0      200612  Net Income (Excl. Extraordinary Items & Good W...
            1      200613  Net Income (Excl. Extraordinary Items & Good W...
            2      200614            Net Income (GAAP) - Surpise, Difference
            3      200615               Net Income (GAAP) - Surpise, Percent
            4      200616        Net Income Normalized - Surpise, Difference
            5      200617           Net Income Normalized - Surpise, Percent
        """
        return cls._dataitems(search=search, package=package)


@_validate_args
def dataitems(search: str = None, package: str = None):
    """
    Usable data items for the estimate data category.

    Parameters
    ----------
        search : str, default None
            | Search word for dataitems name, the search is case-insensitive.

        package : str, default None
            | Search word for package name, the search is case-insensitive.

    Returns
    -------
        pandas.DataFrame
            Data items that belong to cash flow statement data component.

        Columns:
            - *datamodule*
            - *datacomponent*
            - *dataitemid*
            - *datadescription*


    Examples
    --------
        >>> di = prism.estimate.dataitems('net income')
        >>> di[['dataitemid', 'dataitemname']]
            dataitemid                                       dataitemname
        0       200027  Net Income (Excl. Extraordinary Items & Good W...
        1       200028                         Net Income (GAAP) - Actual
        2       200029                     Net Income Normalized - Actual
        3       200203  Net Income (Excl. Extraordinary Items & Good W...
        4       200204  Net Income (Excl. Extraordinary Items & Good W...
        5       200205  Net Income (Excl. Extraordinary Items & Good W...
        6       200206  Net Income (Excl. Extraordinary Items & Good W...
        7       200207  Net Income (Excl. Extraordinary Items & Good W...
        8       200208  Net Income (Excl. Extraordinary Items & Good W...
        9       200209      Net Income (GAAP) - Consensus, # of Estimates
        10      200210                Net Income (GAAP) - Consensus, High
        11      200211                 Net Income (GAAP) - Consensus, Low
        12      200212                Net Income (GAAP) - Consensus, Mean
        13      200213              Net Income (GAAP) - Consensus, Median
    """
    return _list_dataitem(
        datacategoryid=_PrismEstimateComponent.categoryid,
        datacomponentid=None,
        search=search,
        package=package,
    )



import copy
import uuid

from ..._common.const import FrequencyType
from ..._prismcomponent.prismcomponent import _PrismModelComponent, _PrismComponent, _functioncomponent_builder
from ..._utils import _validate_args


_data_category = __name__.split(".")[-1]


class _PrismRiskModelComponent(_PrismModelComponent):
    _component_category_repr  = _data_category

    # def inv(self):
    #     return _functioncomponent_builder("inv", {}, self)

    def __getattribute__(self, name):
        if name not in [
            "get_data",
            "inv",
            "save",
            "copy",
            "query",
            "component_type",
            "categoryid",
            "component_category",
            "componentid",
            "component_name",
            "_query",
            "_dict_to_tree",
            "__class__",
        ]: raise AttributeError(f"{name} not allowed")
        return _PrismModelComponent.__getattribute__(self, name)


class qis(_PrismRiskModelComponent):
    r"""
    | Provides fast and accurate estimators of the covariance matrix based on nonlinear shrinkage for financial assets.
    | Nonlinear shrinkage derived under Frobenius loss and its two cousins, Inverse Stein's loss and Minimum Variance loss, called quadratic-inverse shrinkage (QIS).

    Parameters
    ----------

        frequency : str
            | Desired rebalancing frequency for the risk model
            | Examples:
            |     'M': Generates risk models at the end of every month.
            |     'M-23': Generates risk models on the 23rd day of each month.
            |     'W-M': Generates risk models every Monday of a week.

        period : int
            | Desired number of days (not business days) of data used to calculate a risk model.
            | Please note that the parameter value indicates the time frame covered and not the count of individual data samples.
            | For example, if you input 365 into the parameter, the risk model will be computed using all available data spanning a 365-day period.

        include : bool, default False
            | Determines whether to include pricing data from the rebalancing dates when calculating the risk model. By default, this parameter is set to False

    Returns
    -------
        prism._PrismComponent

    References
    ----------
        Ledoit, O. and Wolf, M. (2022). Quadratic shrinkage for large covariance matrices. Bernoulli.
        Available online at https://www.econ.uzh.ch/en/people/faculty/wolf/publications.html.

    Examples
    --------
        >>> qis = prism.riskmodel.qis("M", 252, False)
        >>> qis.get_data("KR_primary_prism", "2021-01-01", "2021-03-01")
        {'2021-01-31':
        31780818     690828765   682359177   693900517    31780641
        31780818      0.004068    0.000184    0.000325    0.000175    0.000279
        690828765     0.000184    0.001945    0.000134    0.000088    0.000066
        682359177     0.000325    0.000134    0.003783    0.000366    0.000118
        693900517     0.000175    0.000088    0.000366    0.002984   -0.000014
        31780641      0.000279    0.000066    0.000118   -0.000014    0.002087
        ...                ...         ...         ...         ...         ...
        571967822    -0.000072    0.000022    0.000013    0.000032   -0.000020
        31781107      0.000383    0.000077    0.000218    0.000144    0.000156
        681409416     0.000077    0.000031    0.000054    0.000055    0.000035
        683275334     0.000112    0.000056    0.000181    0.000047    0.000059
        1830288132    0.000067    0.000024    0.000114    0.000068    0.000038

                     670416042   650332107   670444531   689898516   133365196  ...
        31780818      0.000403    0.000079    0.000072    0.000782    0.000056  ...
        690828765    -0.000055    0.000043    0.000060    0.000436    0.000134  ...
        682359177     0.000177    0.000419    0.000199    0.000535    0.000429  ...
        693900517     0.000101    0.000297    0.000358    0.000207    0.000228  ...
        31780641      0.000183    0.000090    0.000059    0.000269    0.000051  ...
        ...                ...         ...         ...         ...         ...  ...
        571967822    -0.000041    0.000013    0.000020    0.000052    0.000048  ...
        31781107      0.000175    0.000069    0.000060    0.000305    0.000043  ...
        681409416     0.000037    0.000029    0.000039    0.000074    0.000016  ...
        683275334     0.000077    0.000053    0.000024    0.000224    0.000048  ...
        1830288132    0.000031    0.000042    0.000042    0.000065    0.000031  ...
        }
    """

    @_validate_args
    def __init__(
        self,
        frequency: str,
        sample_period: int,
        include: bool = False,
    ):
        super().__init__(
            frequency=frequency,
            sample_period=sample_period,
            include=include,
        )


def map_attribute(self, attribute):
    attributes = self.attributes[attribute]
    ret = self.copy()
    ret.column = attributes
    ret.index = attributes
    return ret



import inspect
import re
import sys
import pandas as pd
from functools import wraps
from typing import Any, Callable

from prism._utils.exceptions import PrismAuthError, PrismTypeError, PrismValueError
from .._common import const


def handle_jupyter_exception(func):
    @wraps(func)
    def showtraceback(*args, **kwargs):
        # extract exception type, value and traceback
        etype, evalue, tb = sys.exc_info()
        customType = "Prism" in etype.__name__
        if customType:
            print(evalue)
            return
        else:
            # otherwise run the original hook
            value = func(*args, **kwargs)
            return value
    return showtraceback


def handle_sys_exception(exc_type, exc_value, exc_traceback):
    customType = "Prism" in exc_type.__name__
    if customType:
        print(exc_value)
        return
    else:
        # otherwise run the original hook
        return sys.__excepthook__(exc_type, exc_value, exc_traceback)


def _validate_args(function):
    @wraps(function)
    def type_checker(*args, **kwargs):
        func_params_dict = inspect.signature(function).parameters

        # construct param-arg pairs: order matters - do not use set operations
        pos_params = tuple(p for p in func_params_dict.keys() if p not in kwargs.keys())
        if "self" in pos_params:
            pos_params = tuple(p for p in pos_params if p != "self")
            pos_args = dict(zip(pos_params, args[1:]))
        else:
            pos_args = dict(zip(pos_params, args))
        all_args = kwargs.copy()
        all_args.update(pos_args)

        args = list(args)

        if function.__name__ == "fillna":
            if not((all_args.get("value", None) is not None) ^ (all_args.get("method", None) is not None)):
                raise PrismValueError("Must specify one fill 'value' or 'method'.")

        for param, v in func_params_dict.items():

            # if None, not provided (nor required)
            arg = all_args.get(param, v.default)
            if (arg is not None) and (arg is not inspect._empty):
                # check types
                param_type = v.annotation  # if empty, no hints
                if param_type is not inspect._empty:
                    if hasattr(param_type, "__args__"):
                        unionArgs = param_type.__args__
                        if unionArgs is not None and not isinstance(arg, tuple(unionArgs)):
                            raise PrismTypeError(f"Type of {param} is {type(arg)} and not in the Union of {unionArgs}")
                    else:
                        if not issubclass(type(arg), param_type):
                            raise PrismTypeError(f"Type of {param} is {type(arg)} and not {param_type}")

                # validation rules
                arg_enum = None
                if param == "frequency":
                    if function.__name__ == "get_universe":
                        try:
                            arg_enum = const.UniverseFrequencyType(arg)
                        except Exception as e:
                            raise PrismValueError(f"{e}", valid_list=const.UniverseFrequencyType)
                    else:
                        try:
                            arg_enum = const.FrequencyType(arg)
                        except Exception as e:
                            raise PrismValueError(f"{e}", valid_list=const.FrequencyType)
                elif param == "datetype":
                    try:
                        arg_enum = const.DateType(arg)
                    except Exception as e:
                        raise PrismValueError(f"{e}", valid_list=const.DateType)
                elif param == "periodtype":
                    try:
                        arg_enum = const.PeriodType(arg)
                    except Exception as e:
                        raise PrismValueError(f"{e}", valid_list=const.PeriodType)
                elif param == "adjustment": # can be string or bool
                    if arg not in [a.value for a in const.AdjustmentType]:
                        raise PrismValueError(f"Invalid adjustment argument is given. ", valid_list=const.AdjustmentType)
                elif param == "shownid":
                    for idx, id in enumerate(arg):
                        arg[idx] = get_sm_attributevalue(id)
                elif param == "rank_method":
                    try:
                        arg_enum = const.RankType(arg)
                    except Exception as e:
                        raise PrismValueError(f"{e}", valid_list=const.RankType)
                elif param == "method":
                    try:
                        arg_enum = const.FillnaMethodType(arg)
                    except Exception as e:
                        raise PrismValueError(f"{e}", valid_list=const.FillnaMethodType)
                # elif param == "beyond":
                #     try:
                #         assert isinstance(arg, bool)
                #         arg_enum = arg
                #     except Exception as e:
                #         raise PrismValueError(f"{e}", valid_list=BeyondType)
                elif param in ["startdate", "enddate"]:
                    try:
                        date = pd.to_datetime(arg)
                    except Exception as e:
                        raise PrismValueError(f'Unknown string format. Cannot parse "{arg}" to a date')

                    assert (date >= const.BEGINNING_DATE) & (date <= const.ACTIVE_DATE), "Not a valid date."

                elif "universename" in param:
                    regex = re.compile("[@_!#$%^&*()<>?/\|}{~:]`\"'")
                    if isinstance(arg, list):
                        for a in arg:
                            if regex.search(a) is not None:
                                raise PrismValueError("universename not to include special characters")
                    else:
                        if regex.search(arg) is not None:
                            raise PrismValueError("universename not to include special characters")

                elif "min_periods" in param:
                    if arg < 1:
                        raise PrismValueError("min_periods cannot be less than 1")

                elif (function.__name__ == "log") & (param == "base"):
                    if arg <= 0:
                        raise Exception("base condition error")
                elif param == 'preliminary':
                    try:
                        arg_enum = const.FinancialPreliminaryType(arg)
                    except Exception as e:
                        raise PrismValueError(f"{e}", valid_list=const.FinancialPreliminaryType)
                elif param == "setting":
                    from .._common.const import PreferenceType
                    if (arg not in PreferenceType) and (arg != ""):
                        raise PrismValueError(f"Invalid preference keyword is given. ", valid_list=PreferenceType)
                elif param == "settings":
                    from .._common.const import PreferenceType
                    for k in arg.keys():
                        if k not in PreferenceType:
                            raise PrismValueError(f"Invalid preference key is given. ", valid_list=PreferenceType)

                if arg_enum is not None:
                    if param in pos_args.keys():
                        for i, k in enumerate(func_params_dict.keys()):
                            if param == k:
                                break
                        args[i] = arg_enum
                    else:
                        kwargs[param] = arg_enum

        return function(*args, **kwargs)

    return type_checker



def validate_fillna_function(args_dict, function):
    if function.__name__ == "fillna":
        v, m = args_dict.get("value"), args_dict.get("method")
        if not (v is not None) ^ (m is not None):
            raise PrismValueError("Must specify one fill 'value' or 'method'.")


def check_argument_type(arg, param_type, param_name):
    if param_type is inspect.Parameter.empty:
        return

    if isinstance(param_type, type):
        if not isinstance(arg, param_type):
            raise PrismTypeError(
                f"Type of {param_name} is {type(arg)} and not {param_type}")
    elif not any(isinstance(arg, t) for t in param_type.__args__):
        raise PrismTypeError(
            f"Type of {param_name} is {type(arg)} and not in the Union of {param_type.__args__}")


def apply_validation_rules(arg, param_name: str, function_name: str):
    # if param_name equals function name
    if param_name in validation_functions:
        validation_function = validation_functions[param_name]
        return validation_function(arg, function_name)

    # if param_name matches custom pattern
    for _, (pattern_checker, validation_function) in custom_validation_functions.items():
        if pattern_checker(param_name):
            return validation_function(arg)

    # when there is no validation rule to check
    return arg


def frequency_param_validator(arg, function_name):
    if function_name == "get_universe":
        try:
            return const.UniverseFrequencyType(arg)
        except Exception as e:
            raise PrismValueError(
                f"{e}", valid_list=const.UniverseFrequencyType) from e
    else:
        try:
            return const.FrequencyType(arg)
        except Exception as e:
            raise PrismValueError(f"{e}", valid_list=const.FrequencyType) from e


def datetype_param_validator(arg, _):
    try:
        return const.DateType(arg)
    except Exception as e:
        raise PrismValueError(f"{e}", valid_list=const.DateType) from e


def periodtype_param_validator(arg, _):
    try:
        return const.PeriodType(arg)
    except Exception as e:
        raise PrismValueError(f"{e}", valid_list=const.PeriodType) from e


def adjustment_param_validator(arg, _):
    if arg not in [a.value for a in const.AdjustmentType]:
        # need review. original code 'e' is not bounded
        raise PrismValueError(
            'Invalid adjustment argument is given. ', valid_list=const.AdjustmentType)

    return arg


def shownid_param_validator(arg, _):
    for idx, id in enumerate(arg):
        arg[idx] = get_sm_attributevalue(id)

    return arg


def rank_method_param_validator(arg, _):
    try:
        return const.RankType(arg)
    except Exception as e:
        raise PrismValueError(f"{e}", valid_list=const.RankType) from e


def method_param_validator(arg, _):
    try:
        return const.FillnaMethodType(arg)
    except Exception as e:
        raise PrismValueError(f"{e}", valid_list=const.FillnaMethodType) from e


def startdate_param_validator(arg, _):
    try:
        date = pd.to_datetime(arg)
    except Exception as e:
        raise PrismValueError(
            f'Unknown string format. Cannot parse "{arg}" to a date') from e

    assert (date >= const.BEGINNING_DATE) & (
        date <= const.ACTIVE_DATE), "Not a valid date."

    return arg


def enddate_param_validator(arg, _):
    return startdate_param_validator(arg, _)


def base_param_validator(arg, function_name):
    if function_name == "log" and arg <= 0:
        raise Exception("base condition error")

    return arg


def preliminary_param_validator(arg, _):
    try:
        return const.FinancialPreliminaryType(arg)
    except Exception as e:
        raise PrismValueError(
            f"{e}", valid_list=const.FinancialPreliminaryType) from e


def setting_param_validator(arg, _):
    from .._common.const import PreferenceType
    if (arg not in PreferenceType) and (arg != ""):
        raise PrismValueError(
            "Invalid preference keyword is given. ", valid_list=PreferenceType)

    return arg


def settings_param_validator(arg, _):
    from .._common.const import PreferenceType
    for k in arg.keys():
        if k not in PreferenceType:
            raise PrismValueError(
                "Invalid preference key is given. ", valid_list=PreferenceType)

    return arg


# if param_name equals dict's key name
# TODO: refactor this code using eval
validation_functions: dict[str, Callable[[Any, str], Any]] = {
    "frequency": frequency_param_validator,
    "datetype": datetype_param_validator,
    "periodtype": periodtype_param_validator,
    "adjustment": adjustment_param_validator,
    "shownid": shownid_param_validator,
    "rank_method": rank_method_param_validator,
    "method": method_param_validator,
    "startdate": startdate_param_validator,
    "enddate": enddate_param_validator,
    "base": base_param_validator,
    "preliminary": preliminary_param_validator,
    "setting": setting_param_validator,
    "settings": settings_param_validator
}

# custom validators


def custom_universename_validator(arg):
    regex = re.compile("[@_!#$%^&*()<>?/|}{~:]`\"'")
    if isinstance(arg, list):
        for a in arg:
            if regex.search(a) is not None:
                raise PrismValueError(
                    "universename not to include special characters")
    else:
        if regex.search(arg) is not None:
            raise PrismValueError(
                "universename not to include special characters")

    return arg


def custom_min_periods_validator(arg):
    if arg < 1:
        raise PrismValueError("min_periods cannot be less than 1")

    return arg


# uses custum pattern to check param_name
custom_validation_functions: dict[str, tuple[Callable[[str], bool], Callable[[Any], Any]]] = {
    "universename": (lambda param_name: "universename" in param_name, custom_universename_validator),
    "min_periods": (lambda param_name: "min_periods" in param_name, custom_min_periods_validator)
}


def get_sm_attributevalue(attribute: str):
    if const.SMValues is None:
        raise PrismAuthError(f"Please Login First")
    smattributes_lower = {
        a.lower().replace(' ', '').replace('_', ''): const.SMValues[a]
        for a in const.SMValues.keys()
    }
    smattributes_lower.update({
        a.lower().replace(' ', '').replace('_', ''): a
        for a in const.SMValues.values()
    })
    attribute_lower = attribute.lower().replace(' ', '').replace('_', '')
    smattributevalue = smattributes_lower.get(attribute_lower)
    if smattributevalue is None:
        raise PrismValueError(
            f"{attribute} is not a valid Security Master attribute",
            valid_list=list(const.SMValues.keys())
        )
    return smattributevalue


from typing import Union

import pytest

from prism._utils.validate_utils import _validate_args
from prism._utils.exceptions import PrismTypeError


@_validate_args
def add(r: int, l: int):
    return r + l


def test_add():
    assert add(1, 2) == 3
    assert add(1, 5) != 3

    with pytest.raises(PrismTypeError):
        assert add(1, '2') == 3

    with pytest.raises(PrismTypeError):
        assert add(1, 'five') == 6


def test_missing_argument_in_add():
    with pytest.raises((TypeError, PrismTypeError)):
        assert add(1)

    with pytest.raises((TypeError, PrismTypeError)):
        assert add(r=1)

    with pytest.raises((TypeError, PrismTypeError)):
        assert add()

    with pytest.raises((TypeError, PrismTypeError)):
        assert add(1, None)


@_validate_args
def same(r: int, l: float):
    return r, l


def test_same():
    assert same(1, 2.1) == (1, 2.1)
    assert same(1, 5.1) != 3

    # this is intended behavior
    assert same(1, None)
    # assert same(1) # will raise type error


def test_missing_argument_in_same():
    # explicit None can be ignored
    assert same(None, None) == (None, None)
    assert same(1, None) == (1, None)
    assert same(None, 2.0) == (None, 2.0)

    with pytest.raises((TypeError, PrismTypeError)):
        assert same(1)

    with pytest.raises((TypeError, PrismTypeError)):
        assert same(r=1)

    with pytest.raises((TypeError, PrismTypeError)):
        assert same()


@_validate_args
def concat(r: str = '', l: str = ''):
    return r + l


def test_concate():
    assert not concat()
    assert concat('a') == 'a'
    assert concat('a', 'b') == 'ab'
    assert concat(str(1), '2') == '12'

    with pytest.raises(PrismTypeError):
        assert concat('1', 2) == '12'

    with pytest.raises(PrismTypeError):
        assert concat(1, 'five') == '1five'


@_validate_args
def my_union(unionValue: Union[str, float, int]):
    return unionValue


def test_union():
    assert my_union(1) == 1
    assert my_union('abc') == 'abc'
    assert my_union(1.12) == 1.12

    with pytest.raises(PrismTypeError):
        assert not my_union({})

    with pytest.raises(PrismTypeError):
        assert my_union(lambda x: x)



import json
import orjson
import pandas as pd
import pyarrow.parquet as pq
import tarfile
import io
import requests
import tarfile
from zipfile import ZipFile, ZipInfo

import prism
from prism._utils.exceptions import PrismAuthError, PrismError, PrismResponseError
from .._utils import _authentication
from .._common.config import *
from .._common.const import *


def _process_response(res, content=None):
    if res.status_code >= 400:
        if res.status_code == 401:
            raise PrismAuthError(f"Please Login First")

        if res.status_code == 422:
            arg = res.json()["detail"][0]["loc"][-1]
            # enum_list = res.json()["detail"][0]["ctx"]["enum_values"]
            raise PrismResponseError(f'Value at "{arg}" is not a valid enumeration member')

        try:
            err_msg = res.json()["message"]
        except Exception as e:
            if content:
                err_msg = content
            else:
                err_msg = res.content
        raise PrismResponseError(err_msg)
    result = None
    if content:
        resjson = orjson.loads(content)
    else:
        resjson = orjson.loads(res.content)
    restype = resjson["restype"]
    rescontent = resjson["rescontent"]
    if restype == "status":
        result = rescontent
    else:
        data = rescontent.get("data")
        if data is not None:
            if isinstance(data, dict):
                if ("columns" in data.keys()) and (("data" in data.keys())):
                    cols = data.get("columns")
                    data = data.get("data")
                    datecolumns = []
                    cols_to_use = []
                    for c in cols:
                        if c["dtype"] == "datetime64[ns]":
                            datecolumns.append(c["column"])
                        cols_to_use.append(c["column"])
                    result = pd.DataFrame(data, columns=cols_to_use)
                    for c in datecolumns:
                        result[c] = pd.to_datetime(result[c], unit="ns")
                else:
                    result = data
            else:
                result = data
        else:
            result = rescontent
    return result


def unzip_tar(content):
    dfs = {}
    tar_obj = tarfile.open(fileobj=io.BytesIO(content), mode='r')
    fname = tar_obj.getnames()
    metadata = None
    if fname is None: fname = []
    for name in fname:
        f = tar_obj.extractfile(name).read()
        if ".txt" in name:
            metadata = json.loads(f)
        else:
            obj = pq.read_table(io.BytesIO(f))
            dfs[name] = obj.to_pandas()
    return dfs, metadata


def unzip_zip(content):
    dfs = {}
    zip_obj = ZipFile(io.BytesIO(content), mode='r')
    fname = zip_obj.filelist
    if fname is None: fname = []
    metadata = None
    for name in fname:
        f = zip_obj.read(name)
        if ".txt" in name.filename:
            metadata = json.loads(f)
        else:
            obj = pq.read_table(io.BytesIO(f))
            dfs[name.filename] = obj.to_pandas()
    return dfs, metadata


def _process_fileresponse(res, content=None, file_type='tar'):
    if res.status_code >= 400:
        try:
            err_msg = res.json()["message"]
        except:
            err_msg = res.content
        raise PrismError(err_msg)
    try:
        fn = eval(f'unzip_{file_type}')
        return fn(content)
    except Exception as e:
        raise PrismError("Extension response processing error")


def get(url, params=None, stream=False):
    headers = _authentication()
    res = requests.get(url=url, params=params, headers=headers, stream=stream)
    return _process_response(res)


def post(url, params, body, stream=False):
    headers = _authentication()
    res = requests.post(url=url, params=params, json=body, headers=headers, stream=stream)
    return _process_response(res)


def patch(url, params, body):
    headers = _authentication()
    res = requests.patch(url=url, params=params, json=body, headers=headers)
    return _process_response(res)


def delete(url, params=None):
    headers = _authentication()
    res = requests.delete(url=url, params=params, headers=headers)
    return _process_response(res)


def _fetch_and_parse(url, ext, search=None):
    df = get(url, {"search": search})
    if not df.empty:
        directorytype = FILEEXTENSION[ext]
        df["path"] = df["path"].str.split("/", expand=True).replace({None:""}).iloc[:,2:].agg("/".join, axis=1)
        df["path"] = df["path"].str.split("." + ext, expand=True).iloc[:, 0]
        df = df.rename(columns={"path": directorytype + "name"})
        typeid = directorytype + "id"
        if ext == 'ped':
            setattr(df, typeid, getattr(df, typeid).astype('str'))
        else:
            setattr(df, typeid, getattr(df, typeid).astype('Int64'))
    return df[df['folderflag'] == 0].drop("folderflag", axis=1)
